/*

    This file is part of the Maude 3 interpreter.

    Copyright 1997-2003 SRI International, Menlo Park, CA 94025, USA.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

*/

//
//      Forward declarations for AU theory classes.
//
#ifndef _AU_Theory_hh_
#define _AU_Theory_hh_

class AU_Symbol;
class AU_Term;
class AU_ArgumentIterator;
class AU_BaseDagNode;
class AU_DagNode;
class AU_DequeDagNode;
class AU_DagArgumentIterator;
class AU_DequeDagArgumentIterator;
class AU_LhsAutomaton;
class AU_RhsAutomaton;
class AU_Layer;
class AU_Subproblem;
class AU_ExtensionInfo;
class AU_UnificationSubproblem;

#endif
